@@ -1,18 +1,27 @@
 // SPDX-License-Identifier: GPL-3.0-or-later
 use std::path::Path;
 use std::path::PathBuf;
+use std::sync::LazyLock;
 
 use gst::prelude::*;
 use gtk::prelude::*;
 use gtk::subclass::prelude::*;
 use gtk::{gdk, gio, glib, graphene};
-use once_cell::sync::Lazy;
 
+use crate::VideoFormat;
 use crate::ViewfinderState;
+use crate::code_detector::QrCodeDetector;
 
-const BARCODE_TIMEOUT: u32 = 1;
 const PROVIDER_TIMEOUT: u64 = 2;
 
+#[derive(Debug)]
+enum StateChangeState {
+    Equal,
+    Differ,
+    Error,
+    NotDone,
+}
+
 mod imp {
     use std::cell::Cell;
     use std::cell::OnceCell;
@@ -25,23 +34,31 @@ mod imp {
     #[derive(Debug, Default, Properties)]
     #[properties(wrapper_type = super::Viewfinder)]
     pub struct Viewfinder {
-        #[property(get, explicit_notify, builder(Default::default()))]
+        #[property(get, explicit_notify, default)]
         state: Cell<ViewfinderState>,
         #[property(get = Self::detect_codes, set = Self::set_detect_codes, explicit_notify)]
         detect_codes: Cell<bool>,
         #[property(get, set = Self::set_camera, nullable, explicit_notify)]
         camera: RefCell<Option<crate::Camera>>,
+        #[property(get = Self::is_recording, name = "is-recording", type = bool)]
+        pub is_recording_video: RefCell<Option<PathBuf>>,
+        #[property(get, set = Self::set_disable_audio_recording, explicit_notify)]
+        disable_audio_recording: Cell<bool>,
+        #[property(get, set = Self::set_video_format, explicit_notify, default)]
+        video_format: Cell<VideoFormat>,
+        #[property(get, set = Self::set_enable_hw_encoding, explicit_notify)]
+        enable_hw_encoding: Cell<bool>,
 
-        pub zbar_branch: RefCell<Option<gst::Element>>,
+        pub qrcode_branch: RefCell<Option<gst::Element>>,
         pub devices: OnceCell<crate::DeviceProvider>,
         pub camera_src: RefCell<Option<gst::Element>>,
         pub camerabin: OnceCell<gst::Element>,
+        pub camera_element: OnceCell<gst::Element>,
+        pub capsfilter: OnceCell<gst::Element>,
         pub sink_paintable: OnceCell<gst::Element>,
         pub tee: OnceCell<crate::PipelineTee>,
         pub bus_watch: OnceCell<gst::bus::BusWatchGuard>,
 
-        // TODO Port to gio::Task,
-        pub is_recording_video: RefCell<Option<PathBuf>>,
         pub is_stopping_recording: Cell<bool>,
         pub is_taking_picture: Cell<bool>,
         pub is_front_camera: Cell<bool>,
@@ -49,6 +66,7 @@ mod imp {
         pub timeout_handler: RefCell<Option<glib::SourceId>>,
 
         pub picture: gtk::Picture,
+        pub offload: gtk::GraphicsOffload,
     }
 
     impl Viewfinder {
@@ -62,8 +80,12 @@ mod imp {
             }
         }
 
+        fn is_recording(&self) -> bool {
+            self.is_recording_video.borrow().is_some()
+        }
+
         fn detect_codes(&self) -> bool {
-            self.zbar_branch.borrow().is_some()
+            self.qrcode_branch.borrow().is_some()
         }
 
         fn set_detect_codes(&self, value: bool) {
@@ -73,17 +95,17 @@ mod imp {
 
             let tee = self.tee.get().unwrap();
             if value {
-                match create_zbar_bin() {
-                    Ok(zbar_branch) => {
-                        tee.add_branch(&zbar_branch);
-                        self.zbar_branch.replace(Some(zbar_branch));
+                match create_qrcode_bin() {
+                    Ok(qrcode_branch) => {
+                        tee.add_branch(&qrcode_branch);
+                        self.qrcode_branch.replace(Some(qrcode_branch));
                     }
                     Err(err) => {
-                        log::error!("Could not create zbar element: {err}");
+                        log::error!("Could not create qrcode element: {err}");
                     }
                 }
-            } else if let Some(zbar_branch) = self.zbar_branch.take() {
-                tee.remove_branch(&zbar_branch);
+            } else if let Some(qrcode_branch) = self.qrcode_branch.take() {
+                tee.remove_branch(&qrcode_branch);
             }
 
             self.obj().notify_detect_codes();
@@ -138,27 +160,10 @@ mod imp {
             }
 
             if let Some(camera) = camera {
-                // `camera_src` might be `None`, which means the element was
-                // reconfigured and we should keep using it
-
-                // TODO This is incredibly not idiomatic.
-                let guard = self.camera_src.borrow();
-                match camera.source_element(guard.as_ref()) {
-                    Ok(Some((bin, camera_src))) => {
-                        drop(guard);
-                        self.camerabin().set_property("camera-source", &bin);
-                        self.camera_src.replace(Some(camera_src));
-                    }
-                    Ok(None) => (),
-                    Err(err) => {
-                        log::error!("Could not get source element: {err}");
-                        self.set_state(ViewfinderState::Error);
-                        return;
-                    }
+                if let Err(err) = obj.setup_camera_element(&camera) {
+                    log::error!("Could not reconfigure camera element: {err}");
+                    self.set_state(ViewfinderState::Error);
                 }
-
-                let is_front_camera = matches!(camera.location(), crate::CameraLocation::Front);
-                self.is_front_camera.set(is_front_camera);
             }
 
             if obj.is_realized() && matches!(obj.state(), ViewfinderState::Ready) {
@@ -167,6 +172,36 @@ mod imp {
 
             obj.notify_camera();
         }
+
+        fn set_disable_audio_recording(&self, value: bool) {
+            let obj = self.obj();
+
+            if value != self.disable_audio_recording.replace(value) {
+                obj.reset_pipeline();
+                obj.notify_disable_audio_recording();
+            }
+        }
+
+        fn set_video_format(&self, video_format: VideoFormat) {
+            let obj = self.obj();
+
+            if video_format != self.video_format.replace(video_format) {
+                obj.reset_pipeline();
+                obj.notify_video_format();
+            }
+        }
+
+        fn set_enable_hw_encoding(&self, value: bool) {
+            let obj = self.obj();
+
+            if value != self.enable_hw_encoding.replace(value) {
+                match self.video_format.get() {
+                    VideoFormat::Vp8Webm => (),
+                    VideoFormat::H264Mp4 => obj.reset_pipeline(),
+                }
+                obj.notify_enable_hw_encoding();
+            }
+        }
     }
 
     #[glib::object_subclass]
@@ -190,66 +225,108 @@ mod imp {
             let obj = self.obj();
 
             let camerabin = gst::ElementFactory::make("camerabin")
+                .property("location", None::<&str>)
                 .build()
                 .expect("Missing GStreamer Bad Plug-ins");
             self.camerabin.set(camerabin.clone()).unwrap();
 
             let bus = self.camerabin().bus().unwrap();
-            let watch = bus.add_watch_local(
-                glib::clone!(@weak obj => @default-return glib::ControlFlow::Break, move |_, msg| {
-                    obj.on_bus_message(msg);
-                    glib::ControlFlow::Continue
-                }),
-            )
-            .unwrap();
+            let watch = bus
+                .add_watch_local(glib::clone!(
+                    #[weak]
+                    obj,
+                    #[upgrade_or]
+                    glib::ControlFlow::Break,
+                    move |_, msg| {
+                        obj.on_bus_message(msg);
+                        glib::ControlFlow::Continue
+                    }
+                ))
+                .unwrap();
             self.bus_watch.set(watch).unwrap();
 
             let tee = crate::PipelineTee::new();
 
             let paintablesink = gst::ElementFactory::make("gtk4paintablesink")
                 .build()
-                .unwrap();
+                .expect("Missing gst-plugin-gtk4");
 
             let paintable = paintablesink.property::<gdk::Paintable>("paintable");
-            let is_gl_supported = paintable
-                .property::<Option<gdk::GLContext>>("gl-context")
-                .is_some();
-            let sink = if is_gl_supported {
-                gst::ElementFactory::make("glsinkbin")
-                    .property("sink", &paintablesink)
-                    .build()
-                    .expect("Missing GStreamer Base Plug-ins")
-            } else {
+
+            let is_yuv_natively_supported = {
+                let yuv_caps =
+                    gst_video::video_make_raw_caps(&[gst_video::VideoFormat::Yuy2]).build();
+                !paintablesink
+                    .pad_template("sink")
+                    .unwrap()
+                    .caps()
+                    .intersect(&yuv_caps)
+                    .is_empty()
+            };
+            let sink = if is_yuv_natively_supported {
                 let bin = gst::Bin::default();
-                let convert = gst::ElementFactory::make("videoconvert")
-                    .build()
-                    .expect("Missing GStreamer Base Plug-ins");
 
-                bin.add(&convert).unwrap();
                 bin.add(&paintablesink).unwrap();
-                convert.link(&paintablesink).unwrap();
-
                 bin.add_pad(
-                    &gst::GhostPad::with_target(&convert.static_pad("sink").unwrap()).unwrap(),
+                    &gst::GhostPad::with_target(&paintablesink.static_pad("sink").unwrap())
+                        .unwrap(),
                 )
                 .unwrap();
 
                 bin.upcast()
+            } else {
+                let is_gl_supported = paintable
+                    .property::<Option<gdk::GLContext>>("gl-context")
+                    .is_some();
+                if is_gl_supported {
+                    gst::ElementFactory::make("glsinkbin")
+                        .property("sink", &paintablesink)
+                        .build()
+                        .expect("Missing GStreamer Base Plug-ins")
+                } else {
+                    let bin = gst::Bin::default();
+                    let convert = gst::ElementFactory::make("videoconvert")
+                        .build()
+                        .expect("Missing GStreamer Base Plug-ins");
+
+                    bin.add(&convert).unwrap();
+                    bin.add(&paintablesink).unwrap();
+                    convert.link(&paintablesink).unwrap();
+
+                    bin.add_pad(
+                        &gst::GhostPad::with_target(&convert.static_pad("sink").unwrap()).unwrap(),
+                    )
+                    .unwrap();
+
+                    bin.upcast()
+                }
             };
 
             tee.add_branch(&sink);
             camerabin.set_property("viewfinder-sink", &tee);
 
+            let videoconvert_video = gst::ElementFactory::make("videoconvert")
+                .build()
+                .expect("Missing GStreamer Base Plug-ins");
+            camerabin.set_property("video-filter", &videoconvert_video);
+
+            let videoconvert_image = gst::ElementFactory::make("videoconvert")
+                .build()
+                .expect("Missing GStreamer Base Plug-ins");
+            camerabin.set_property("image-filter", &videoconvert_image);
+
             self.sink_paintable.set(paintablesink).unwrap();
 
             self.picture
                 .set_accessible_role(gtk::AccessibleRole::Presentation);
             self.picture.set_hexpand(true);
             self.picture.set_vexpand(true);
-            self.picture.set_content_fit(gtk::ContentFit::Fill);
-            self.picture.set_parent(&*obj);
             self.picture.set_paintable(Some(&paintable));
 
+            self.offload.set_child(Some(&self.picture));
+            self.offload.set_parent(&*obj);
+            self.offload.set_black_background(true);
+
             self.tee.set(tee).unwrap();
 
             let devices = crate::DeviceProvider::instance();
@@ -259,33 +336,49 @@ mod imp {
             if devices.started() {
                 obj.init();
             } else {
-                devices.connect_started_notify(glib::clone!(@weak obj => move |_| {
-                    obj.init();
-                }));
+                devices.connect_started_notify(glib::clone!(
+                    #[weak]
+                    obj,
+                    move |_| {
+                        obj.init();
+                    }
+                ));
             }
 
-            devices.connect_camera_added(glib::clone!(@weak obj => move |_, camera| {
-                if matches!(obj.state(), ViewfinderState::NoCameras | ViewfinderState::Loading | ViewfinderState::Error) {
-                    obj.imp().set_state(ViewfinderState::Ready);
-                    obj.set_camera(Some(camera.clone()));
-                }
-            }));
-
-            devices.connect_camera_removed(glib::clone!(@weak obj => move |devices, camera| {
-                let imp = obj.imp();
-                if Some(camera) == imp.camera.borrow().as_ref() {
-                    obj.cancel_current_operation();
-
-                    let next_camera = devices.camera(0);
-                    let is_none = next_camera.is_none();
-                    obj.set_camera(next_camera);
-                    if is_none {
-                        obj.imp().set_state(ViewfinderState::NoCameras);
+            devices.connect_camera_added(glib::clone!(
+                #[weak]
+                obj,
+                move |_, camera| {
+                    if matches!(
+                        obj.state(),
+                        ViewfinderState::NoCameras
+                            | ViewfinderState::Loading
+                            | ViewfinderState::Error
+                    ) {
+                        obj.imp().set_state(ViewfinderState::Ready);
+                        obj.set_camera(Some(camera.clone()));
                     }
                 }
-            }));
+            ));
+
+            devices.connect_camera_removed(glib::clone!(
+                #[weak]
+                obj,
+                move |devices, camera| {
+                    let imp = obj.imp();
+                    if Some(camera) == imp.camera.borrow().as_ref() {
+                        obj.cancel_current_operation();
+
+                        let next_camera = devices.camera(0);
+                        let is_none = next_camera.is_none();
+                        obj.set_camera(next_camera);
+                        if is_none {
+                            obj.imp().set_state(ViewfinderState::NoCameras);
+                        }
+                    }
+                }
+            ));
 
-            log::debug!("Setup recording");
             obj.setup_recording();
         }
 
@@ -299,11 +392,11 @@ mod imp {
                 log::error!("Could not stop camerabin: {err}");
             }
 
-            self.picture.unparent();
+            self.offload.unparent();
         }
 
         fn signals() -> &'static [glib::subclass::Signal] {
-            static SIGNALS: Lazy<Vec<glib::subclass::Signal>> = Lazy::new(|| {
+            static SIGNALS: LazyLock<Vec<glib::subclass::Signal>> = LazyLock::new(|| {
                 vec![
                     // These are emitted whenever the saving process finishes,
                     // successful or not.
@@ -314,7 +407,7 @@ mod imp {
                         .param_types([Option::<gio::File>::static_type()])
                         .build(),
                     glib::subclass::Signal::builder("code-detected")
-                        .param_types([crate::CodeType::static_type(), String::static_type()])
+                        .param_types([glib::Bytes::static_type()])
                         .build(),
                 ]
             });
@@ -327,32 +420,24 @@ mod imp {
             self.parent_realize();
 
             if matches!(self.obj().state(), ViewfinderState::Ready) {
+                log::debug!("Viewfinder realized: starting stream");
                 self.obj().start_stream();
             }
         }
 
         fn unrealize(&self) {
+            log::debug!("Viewfinder unrealized: stopping stream");
             self.obj().stop_stream();
 
             self.parent_unrealize();
         }
 
         fn snapshot(&self, snapshot: &gtk::Snapshot) {
-            let w = self.obj().width() as f32;
-
-            // This is the composition of translate (-w / 2.0, 0.0), map x to
-            // -x, and translate (w / 2.0 , 0.0). Note that gsk matrices are
-            // transposed (they act on row vectors).
             if self.is_front_camera.get() {
-                #[rustfmt::skip]
-                let flip_matrix = graphene::Matrix::from_float([
-                    -1.0,  0.0,  0.0,  0.0,
-                     0.0,  1.0,  0.0,  0.0,
-                     0.0,  0.0,  1.0,  0.0,
-                       w,  0.0,  0.0,  1.0,
-                ]);
+                let w = self.obj().width() as f32;
                 snapshot.save();
-                snapshot.transform_matrix(&flip_matrix);
+                snapshot.translate(&graphene::Point::new(w, 0.0));
+                snapshot.scale(-1.0, 1.0);
                 self.parent_snapshot(snapshot);
                 snapshot.restore();
             } else {
@@ -390,6 +475,18 @@ glib::wrapper! {
     ///
     ///  Readable | Writable
     ///
+    /// ### `video-format`
+    /// The video format for recordings. `[crate::is_h264_encoding_supported]`
+    /// can be used to detect whether there is h264 support.
+    ///
+    ///  Readable | Writable
+    ///
+    /// ### `enable-hw-encoding`
+    /// Whether to enable hardware video encoding.
+    /// `[crate::is_hardware_encoding_supported]` can be used to detect whether
+    /// the system supports hardware encoding for a given format.
+    ///
+    ///  Readable | Writable
     ///
     /// #### `camera`
     ///  The camera that is currently being used.
@@ -427,7 +524,8 @@ glib::wrapper! {
     ///
     /// [`gtk::prelude::WidgetExt`][trait@gtk::prelude::WidgetExt], [`glib::ObjectExt`][trait@gtk::glib::ObjectExt]
     pub struct Viewfinder(ObjectSubclass<imp::Viewfinder>)
-        @extends gtk::Widget;
+        @extends gtk::Widget,
+        @implements gtk::Accessible, gtk::Buildable, gtk::ConstraintTarget;
 }
 
 impl Default for Viewfinder {
@@ -527,8 +625,13 @@ impl Viewfinder {
         }
 
         // Set after we cannot fail anymore.
-        imp.is_recording_video
-            .replace(Some(location.as_ref().to_owned()));
+        if imp
+            .is_recording_video
+            .replace(Some(location.as_ref().to_owned()))
+            .is_none_or(|old| old != location.as_ref())
+        {
+            self.notify_is_recording();
+        };
 
         let camerabin = imp.camerabin();
         camerabin.set_property_from_str("mode", "mode-video");
@@ -568,15 +671,6 @@ impl Viewfinder {
         Ok(())
     }
 
-    /// Gets whether a recording is in progress.
-    ///
-    /// # Returns
-    ///
-    /// if a recording is in progress.
-    pub fn is_recording(&self) -> bool {
-        self.imp().is_recording_video.borrow().is_some()
-    }
-
     pub fn connect_picture_done<F: Fn(&Self, Option<&gio::File>) + 'static>(&self, f: F) {
         self.connect_closure(
             "picture-done",
@@ -597,24 +691,83 @@ impl Viewfinder {
         );
     }
 
-    pub fn connect_code_detected<F: Fn(&Self, crate::CodeType, &str) + 'static>(&self, f: F) {
+    pub fn connect_code_detected<F: Fn(&Self, glib::Bytes) + 'static>(&self, f: F) {
         self.connect_closure(
             "code-detected",
             false,
-            glib::closure_local!(|obj, data_type, data| {
-                f(obj, data_type, data);
+            glib::closure_local!(|obj, data| {
+                f(obj, data);
             }),
         );
     }
 
     /// Starts the viewfinder.
     pub fn start_stream(&self) {
-        // TODO Present a toast, banner, or message dialog with the error.
-        if let Err(err) = self.imp().camerabin().set_state(gst::State::Playing) {
-            log::error!("Could not start camerabin: {err}");
-            self.imp().set_state(ViewfinderState::Error);
-        } else {
-            log::debug!("Camerabin state succesfully set to PLAYING");
+        glib::spawn_future_local(glib::clone!(
+            #[weak(rename_to = obj)]
+            self,
+            async move {
+                obj.change_state_inner(gst::State::Playing).await;
+            }
+        ));
+    }
+
+    // It is not needed to call this for gst::State::Null.
+    async fn change_state_inner(&self, state: gst::State) {
+        let (sender, receiver) = futures_channel::oneshot::channel();
+
+        let camerabin = self.imp().camerabin();
+        std::thread::spawn(glib::clone!(#[weak] camerabin, move || {
+            let timeout = gst::format::ClockTime::from_seconds(2);
+            let (res, current_state, pending_state) = camerabin.state(Some(timeout));
+            let new_state_is = match res {
+                Ok(change_done) => {
+                    if change_done == gst::StateChangeSuccess::Async {
+                        camerabin.set_locked_state(true);
+                        log::debug!("Camerabin could not change its state from {current_state:?} to {pending_state:?}");
+                        StateChangeState::NotDone
+                    } else if current_state == state {
+                        StateChangeState::Equal
+                    } else {
+                        StateChangeState::Differ
+                    }
+                }
+                Err(err) => {
+                    log::error!("Previous camerabin state changed failed: {err}");
+                    StateChangeState::Error
+                }
+            };
+            sender.send(new_state_is).unwrap();
+        }))
+            .join()
+            .unwrap();
+
+        let change_state = receiver.await.unwrap();
+        match change_state {
+            StateChangeState::Equal => {
+                // Nothing to do, the new state matches the current one.
+            }
+            StateChangeState::NotDone => {
+                log::debug!("Aborting camerabin state change {state:?}");
+                camerabin.abort_state();
+                camerabin.set_locked_state(false);
+                self.set_camerabin_state(state);
+            }
+            // If the previous state change failed, we might as well try to set it now.
+            StateChangeState::Error | StateChangeState::Differ => self.set_camerabin_state(state),
+        }
+    }
+
+    fn set_camerabin_state(&self, state: gst::State) {
+        match self.imp().camerabin().set_state(state) {
+            Err(err) => {
+                log::error!("Could not start camerabin: {err}");
+                self.imp().set_state(ViewfinderState::Error);
+            }
+            Ok(gst::StateChangeSuccess::Async) => {
+                log::debug!("Trying to set camerabin state to {state:?}");
+            }
+            Ok(_) => log::debug!("Camerabin successfully state set to {state:?}"),
         }
     }
 
@@ -626,7 +779,7 @@ impl Viewfinder {
             log::error!("Could not pause camerabin: {err}");
             self.imp().set_state(ViewfinderState::Error);
         } else {
-            log::debug!("Camerabin state succesfully set to NULL");
+            log::debug!("Camerabin state successfully set to NULL");
         }
     }
 
@@ -643,12 +796,10 @@ impl Viewfinder {
                 Some(s) if s.has_name("video-done") => {
                     self.on_video_done();
                 }
-                Some(s) if s.has_name("barcode") => {
-                    let type_str = s.get::<&str>("type").unwrap();
-                    let data_type = type_str.into();
-                    let data = s.get::<&str>("symbol").unwrap();
+                Some(s) if s.has_name("qrcode") => {
+                    let data = s.get::<glib::Bytes>("payload").unwrap();
 
-                    self.on_barcode_detected(data_type, data);
+                    self.emit_code_detected(data);
                 }
                 _ => (),
             },
@@ -666,25 +817,12 @@ impl Viewfinder {
         self.imp().is_stopping_recording.set(false);
 
         if let Some(path) = self.imp().is_recording_video.take() {
+            self.notify_is_recording();
             let file = gio::File::for_path(path);
             self.emit_recording_done(Some(&file));
         }
     }
 
-    fn on_barcode_detected(&self, data_type: crate::CodeType, data: &str) {
-        // We don't emit the signal if we just emited it
-        if self.imp().timeout_handler.borrow().is_none() {
-            let id = glib::timeout_add_seconds_local_once(
-                BARCODE_TIMEOUT,
-                glib::clone!(@weak self as obj => move || {
-                    obj.imp().timeout_handler.take();
-                }),
-            );
-            self.imp().timeout_handler.replace(Some(id));
-            self.emit_code_detected(data_type, data);
-        }
-    }
-
     fn on_pipeline_error(&self, err: &gst::message::Error) {
         log::error!(
             "Bus Error from {:?}\n{}\n{:?}",
@@ -707,6 +845,7 @@ impl Viewfinder {
             self.emit_picture_done(None);
         }
         if imp.is_recording_video.replace(None).is_some() {
+            self.notify_is_recording();
             self.emit_recording_done(None);
         }
         imp.is_stopping_recording.set(false);
@@ -720,8 +859,9 @@ impl Viewfinder {
         self.emit_by_name::<()>("recording-done", &[&file]);
     }
 
-    fn emit_code_detected(&self, data_type: crate::CodeType, data: &str) {
-        self.emit_by_name::<()>("code-detected", &[&data_type, &data]);
+    fn emit_code_detected(&self, data: glib::Bytes) {
+        log::info!("Code detected: {}", String::from_utf8_lossy(&data));
+        self.emit_by_name::<()>("code-detected", &[&data]);
     }
 
     fn set_tags(&self) {
@@ -751,27 +891,128 @@ impl Viewfinder {
     fn setup_recording(&self) {
         use gst_pbutils::encoding_profile::EncodingProfileBuilder;
 
-        let video_profile =
-            gst_pbutils::EncodingVideoProfile::builder(&gst::Caps::builder("video/x-vp8").build())
+        let profile = match self.video_format() {
+            VideoFormat::H264Mp4 => {
+                let mut hw_encoder_found = false;
+                let registry = gst::Registry::get();
+                if let Some(encoder) = registry.lookup_feature("vah264lpenc") {
+                    if self.enable_hw_encoding() {
+                        encoder.set_rank(gst::Rank::PRIMARY + 2);
+                    } else {
+                        encoder.set_rank(gst::Rank::NONE);
+                    }
+                    hw_encoder_found = true;
+                }
+                if let Some(encoder) = registry.lookup_feature("vah264enc") {
+                    if self.enable_hw_encoding() {
+                        encoder.set_rank(gst::Rank::PRIMARY + 1);
+                    } else {
+                        encoder.set_rank(gst::Rank::NONE);
+                    }
+                    hw_encoder_found = true;
+                }
+                if let Some(encoder) = registry.lookup_feature("v4l2h264enc") {
+                    if self.enable_hw_encoding() {
+                        encoder.set_rank(gst::Rank::PRIMARY + 1);
+                    } else {
+                        encoder.set_rank(gst::Rank::NONE);
+                    }
+                    hw_encoder_found = true;
+                }
+                log::debug!(
+                    "Setting up recording with h264/mp4 profile {} hw acceleration",
+                    if self.enable_hw_encoding() && hw_encoder_found {
+                        "with"
+                    } else {
+                        "without"
+                    }
+                );
+
+                let caps = gst::Caps::builder("video/quicktime").build();
+                let mut container_profile = gst_pbutils::EncodingContainerProfile::builder(&caps)
+                    .name("MP4 audio/video")
+                    .description("Standard MP4/H264/MP3");
+
+                let video_profile = gst_pbutils::EncodingVideoProfile::builder(
+                    &gst::Caps::builder("video/x-h264").build(),
+                )
+                .variable_framerate(true)
+                .build();
+                container_profile = container_profile.add_profile(video_profile);
+
+                if !self.disable_audio_recording() {
+                    let audio_profile = gst_pbutils::EncodingAudioProfile::builder(
+                        &gst::Caps::builder("audio/mpeg").build(),
+                    )
+                    .build();
+                    container_profile = container_profile.add_profile(audio_profile);
+                }
+
+                container_profile.build()
+            }
+            VideoFormat::Vp8Webm => {
+                let mut hw_encoder_found = false;
+                let registry = gst::Registry::get();
+                if let Some(encoder) = registry.lookup_feature("vavp8lpenc") {
+                    if self.enable_hw_encoding() {
+                        encoder.set_rank(gst::Rank::PRIMARY + 2);
+                    } else {
+                        encoder.set_rank(gst::Rank::NONE);
+                    }
+                    hw_encoder_found = true;
+                }
+                if let Some(encoder) = registry.lookup_feature("vavp8enc") {
+                    if self.enable_hw_encoding() {
+                        encoder.set_rank(gst::Rank::PRIMARY + 1);
+                    } else {
+                        encoder.set_rank(gst::Rank::NONE);
+                    }
+                    hw_encoder_found = true;
+                }
+                if let Some(encoder) = registry.lookup_feature("v4l2vp8enc") {
+                    if self.enable_hw_encoding() {
+                        encoder.set_rank(gst::Rank::PRIMARY + 1);
+                    } else {
+                        encoder.set_rank(gst::Rank::NONE);
+                    }
+                    hw_encoder_found = true;
+                }
+                log::debug!(
+                    "Setting up recording with vp8/webm profile {} hw acceleration",
+                    if self.enable_hw_encoding() && hw_encoder_found {
+                        "with"
+                    } else {
+                        "without"
+                    }
+                );
+
+                let caps = gst::Caps::builder("video/webm").build();
+                let mut container_profile = gst_pbutils::EncodingContainerProfile::builder(&caps)
+                    .name("WebM audio/video")
+                    .description("Standard WebM/VP8/Vorbis");
+
+                let video_profile = gst_pbutils::EncodingVideoProfile::builder(
+                    &gst::Caps::builder("video/x-vp8").build(),
+                )
                 .preset("Profile Realtime")
                 .variable_framerate(true)
                 .build();
-        let audio_profile = gst_pbutils::EncodingAudioProfile::builder(
-            &gst::Caps::builder("audio/x-vorbis").build(),
-        )
-        .build();
-        let profiles = gst_pbutils::EncodingContainerProfile::builder(
-            &gst::Caps::builder("video/webm").build(),
-        )
-        .name("WebM audio/video")
-        .description("Standard WebM/VP8/Vorbis")
-        .add_profile(video_profile)
-        .add_profile(audio_profile)
-        .build();
+                container_profile = container_profile.add_profile(video_profile);
+
+                if !self.disable_audio_recording() {
+                    let audio_profile = gst_pbutils::EncodingAudioProfile::builder(
+                        &gst::Caps::builder("audio/x-vorbis").build(),
+                    )
+                    .build();
+                    container_profile = container_profile.add_profile(audio_profile);
+                }
+
+                container_profile.build()
+            }
+        };
 
         let camerabin = self.imp().camerabin();
-
-        camerabin.set_property("video-profile", profiles);
+        camerabin.set_property("video-profile", profile);
     }
 
     fn init(&self) {
@@ -790,26 +1031,161 @@ impl Viewfinder {
 
         glib::timeout_add_local_once(
             std::time::Duration::from_secs(PROVIDER_TIMEOUT),
-            glib::clone!(@weak self as obj => move || {
-                if matches!(obj.state(), ViewfinderState::Loading) {
-                    obj.imp().set_state(ViewfinderState::NoCameras);
+            glib::clone!(
+                #[weak(rename_to = obj)]
+                self,
+                move || {
+                    if matches!(obj.state(), ViewfinderState::Loading) {
+                        obj.imp().set_state(ViewfinderState::NoCameras);
+                    }
                 }
-            }),
+            ),
         );
     }
+
+    fn create_camera_element(
+        &self,
+        device_src: &gst::Element,
+    ) -> Result<gst::Element, glib::BoolError> {
+        use gst::prelude::*;
+
+        let bin = gst::Bin::new();
+
+        let capsfilter = gst::ElementFactory::make("capsfilter").build()?;
+        let decodebin3 = gst::ElementFactory::make("decodebin3").build()?;
+        let capsfilter_post_decode = gst::ElementFactory::make("capsfilter").build()?;
+        let caps_post_decode = gst::Caps::builder("video/x-raw").build();
+        capsfilter_post_decode.set_property("caps", &caps_post_decode);
+
+        bin.add_many([
+            device_src,
+            &capsfilter,
+            &decodebin3,
+            &capsfilter_post_decode,
+        ])?;
+        gst::Element::link_many([device_src, &capsfilter, &decodebin3])?;
+
+        self.imp().capsfilter.set(capsfilter).unwrap();
+
+        let (sender, receiver) = futures_channel::oneshot::channel::<bool>();
+        let sender = std::sync::Arc::new(std::sync::Mutex::new(Some(sender)));
+        decodebin3.connect_pad_added(glib::clone!(#[weak] capsfilter_post_decode, move |_, pad| {
+            if pad.stream().is_some_and(|stream| matches!(stream.stream_type(), gst::StreamType::VIDEO)) {
+                let has_succeeded = pad.link(&capsfilter_post_decode.static_pad("sink").unwrap())
+                                       .inspect_err(|err| {
+                                           log::error!("Failed to link decodebin3:video_%u pad with capsfilter_post_decode:sink pad: {err}");
+                                       })
+                                       .is_ok();
+                let mut guard = sender.lock().unwrap();
+                if let Some(sender) = guard.take() {
+                    let _ = sender.send(has_succeeded);
+                }
+            }
+        }));
+
+        glib::spawn_future_local(glib::clone!(
+            #[weak(rename_to = viewfinder)]
+            self,
+            async move {
+                let has_succeeded = receiver.await.unwrap_or_default();
+                if !has_succeeded {
+                    viewfinder.imp().set_state(ViewfinderState::Error);
+                }
+            }
+        ));
+
+        let pad = capsfilter_post_decode.static_pad("src").unwrap();
+        let ghost_pad = gst::GhostPad::with_target(&pad)?;
+        ghost_pad.set_active(true)?;
+
+        bin.add_pad(&ghost_pad)?;
+
+        let wrappercamerabinsrc = gst::ElementFactory::make("wrappercamerabinsrc")
+            .property("video-source", &bin)
+            .build()
+            .expect("Missing GStreamer Bad Plug-ins");
+
+        Ok(wrappercamerabinsrc)
+    }
+
+    fn setup_camera_element(&self, camera: &crate::Camera) -> Result<(), glib::BoolError> {
+        let imp = self.imp();
+
+        if let Some(element) = imp.camera_element.get() {
+            camera.reconfigure(element)?;
+        } else {
+            let element = camera.create_element()?;
+
+            let wrapper = self.create_camera_element(&element)?;
+            imp.camerabin().set_property("camera-source", &wrapper);
+
+            imp.camera_element.set(element).unwrap();
+        }
+
+        if let Some(capsfilter) = imp.capsfilter.get() {
+            let caps = camera.best_caps();
+            capsfilter.set_property("caps", &caps);
+        }
+
+        let is_front_camera = !matches!(camera.location(), crate::CameraLocation::Back);
+        imp.is_front_camera.set(is_front_camera);
+
+        Ok(())
+    }
+
+    fn reset_pipeline(&self) {
+        if matches!(
+            self.imp().camerabin().current_state(),
+            gst::State::Playing | gst::State::Paused
+        ) {
+            self.stop_stream();
+            self.setup_recording();
+            self.start_stream();
+        } else {
+            self.setup_recording();
+        }
+    }
 }
 
-fn create_zbar_bin() -> Result<gst::Element, glib::BoolError> {
+fn create_qrcode_bin() -> Result<gst::Element, glib::BoolError> {
     let bin = gst::Bin::new();
 
+    let videorate = gst::ElementFactory::make("videorate").build()?;
+    videorate.set_property("max-rate", 5);
+    videorate.set_property("drop-only", true);
     let videoconvert = gst::ElementFactory::make("videoconvert").build()?;
-    let zbar = gst::ElementFactory::make("zbar").build()?;
+
+    // Ensure a copy is made
+    let capsfilter = gst::ElementFactory::make("capsfilter").build()?;
+    let caps = gst::Caps::builder("video/x-raw")
+        .field("format", gst_video::VideoFormat::Gray8.to_str())
+        .build();
+    capsfilter.set_property("caps", &caps);
+
+    let queue = gst::ElementFactory::make("queue").build()?;
+    let qrcode = QrCodeDetector::new().upcast::<gst::Element>();
     let fakesink = gst::ElementFactory::make("fakesink").build()?;
 
-    bin.add_many([&videoconvert, &zbar, &fakesink]).unwrap();
-    gst::Element::link_many([&videoconvert, &zbar, &fakesink]).unwrap();
+    bin.add_many([
+        &videorate,
+        &videoconvert,
+        &capsfilter,
+        &queue,
+        &qrcode,
+        &fakesink,
+    ])
+    .unwrap();
+    gst::Element::link_many([
+        &videorate,
+        &videoconvert,
+        &capsfilter,
+        &queue,
+        &qrcode,
+        &fakesink,
+    ])
+    .unwrap();
 
-    let pad = videoconvert.static_pad("sink").unwrap();
+    let pad = videorate.static_pad("sink").unwrap();
     let ghost_pad = gst::GhostPad::with_target(&pad).unwrap();
     ghost_pad.set_active(true).unwrap();
     bin.add_pad(&ghost_pad).unwrap();
