@@ -41,8 +41,8 @@ const PAGE_SPACING_FIXED: f32 = 25.0;
 const PAGE_SPACING_PERCENT: f32 = 0.02;
 
 mod imp {
-    use glib::subclass::Signal;
     use glib::Properties;
+    use glib::subclass::Signal;
 
     use super::*;
 
@@ -264,7 +264,7 @@ mod imp {
 glib::wrapper! {
     pub struct SlidingView(ObjectSubclass<imp::SlidingView>)
         @extends gtk::Widget,
-        @implements gtk::Buildable, adw::Swipeable;
+        @implements gtk::Buildable, adw::Swipeable, gtk::Accessible, gtk::ConstraintTarget;
 }
 
 impl SlidingView {
@@ -458,6 +458,7 @@ impl SlidingView {
     fn shift_position(&self, old_index: Option<usize>) {
         if let (Some(old_index), Some(new_index)) = (old_index, self.current_index()) {
             let shift = new_index as f64 - old_index as f64;
+            // TODO Use Cell::update once stabilized.
             self.imp().position_shift.set(self.position_shift() + shift);
         }
     }
